import { CommHmiNodeServer, Cnc } from "./libfagorclient.es5.js";

/**
 * This is just a sample code showing how to use lfc API.
 *
 * No good practices are being specially followed.
 *
 */
const conn = new CommHmiNodeServer(); // Implements the connection process with the main HMI Node Server
const cnc = new Cnc(conn); // CNC class
window["lfc"] = {
  conn,
  cnc,
};
log("Main Objects initialized");

const PORT = 3000;

let baseData = {
  host: "",
  user: "",
  pass: "",
  connect: false,
};

// Fill up form values if recorded?
const credentials = localStorage.getItem("__lfcCredencial");
if (credentials) {
  baseData = { ...baseData, ...JSON.parse(credentials) };
}

// Connection form submit hook
$("#connectForm").on("submit", (e) => {
  e.preventDefault();

  baseData.host = $("#hostField").val();
  baseData.user = $("#userField").val();
  baseData.pass = $("#passField").val();
  baseData.connect = $("#autoField").prop("checked");

  localStorage.setItem("__lfcCredencial", JSON.stringify(baseData));

  conn.connect({
    host: baseData.host,
    port: PORT,
    username: baseData.user,
    password: baseData.pass,
  });
});

$("#hostField").val(baseData.host);
$("#userField").val(baseData.user);
$("#passField").val(baseData.pass);
$("#autoField").attr("checked", baseData.connect);

if (baseData.connect) {
  $("#connectForm").trigger("submit");
}

const currentAxes = [],
  currentSpindles = [];
const $tbody = $("#chn1Table tbody");
const $tSpindlesbody = $("#chn1SpindlesTable tbody");

let pollingTimer = null,
  axesUpdatesSubscribed = false,
  spindlesUpdatesSubscribed = false,
  first = true;

// online event, all services and properties are reachable and usable.
// It is recommended that every operation that involves an interaction with cnc be executed in online state to
// avoid unexpected exceptions.

// In every reconnection we remove all cnc event listeners, like cnc model is requested every reconnection. That's because
// when libfagorclient lose connection with cnc we cannot ensure that data stored actually corresponds with reality.

conn.on("online", () => {
  log(`Connection established to ${baseData.host}:${PORT}`);
  $("#connectForm").slideUp();

  if (first) {
    log("Requesting model loop [cnc.requestModel()]");
    cnc.requestModel();
    first = false;
  }
});

// This callback will be called once.
// cncModel will contain a simple **plain** JS Object
cnc.on("created", function (cncModel) {
  // We are going to keep current axes on this array
  currentAxes.length = 0;
  $tbody.html("");
  // We are going to keep current spindles on this array
  $tSpindlesbody.html("");
  currentSpindles.length = 0;

  clearInterval(pollingTimer);

  connectGraphics();

  //we must remove all active listeners to avoid memory leaks.
  unsubscribeAxesSpindlesUpdates();

  // We choose a pooling in order to update cnc values
  // an "updated" event is send on cnc, but this approach is easier.
  pollingTimer = setInterval(() => {
    currentAxes.map((axLogicId) => {
      const ax = cnc.getAxis(axLogicId);
      if (ax) {
        const $tr = $(`#ax${axLogicId}`);
        $(".name", $tr).html(ax.name);
        $(".atippos", $tr).html(ax.atippos.external.toFixed(3));
        $(".home", $tr).html(ax.pos.external.toFixed(3));
        $(".parked", $tr).html(ax.parked);
      }
    });

    currentSpindles.map((spLogicId) => {
      const sp = cnc.getSpindle(spLogicId);

      if (sp) {
        const $tr = $(`#sp${spLogicId}`);
        $(".name", $tr).html(sp.name);
        $(".pos", $tr).html(sp.pos.external.toFixed(3));
        $(".sreal", $tr).html(sp.sreal.external.toFixed(3));
        $(".parked", $tr).html(sp.parked);
      }
    });
  }, 60);

  // cncModel is not a living Object; just an inmutable plain JS object
  // a new reference will be pushed when a value changed
  updateAxes(cncModel.getChannel(1).axes);
  updateSpindles(cncModel.getChannel(1).spindles);

  subscribeToAxesSpindlesUpdates();

  requestCncStatusAndPaintIt();
});

// EVENTS
// When Cnc state event arrived:
cnc.on("CncStateEvent", (event) => {
  $(".status").html(event.data.state);
});

// When Cnc mode event arrived:
cnc.on("CncModeEvent", (event) => {
  log(`Cnc operation mode has changed: ${event.data.mode}`);
  $(".mode").html(event.data.mode);
  //Mode has changed, we update the button styles
  updateBlockModeButtons(event.data.mode);
});

// To know which have been disconnected
conn.on("offline", (link) => {
  log(`Connection Lost: ${JSON.stringify(link)}`);
  $("#connectForm").slideDown();
});

$(".blockButton").on("click", async function () {
  const id = $(this).attr("id");

  const channel = cnc.getChannel("1");
  if (id === "enter") {
    log(`About to execute [cnc.getChannel('1').cncServices.enterBlockMode()]`);
    channel.cncServices.enterBlockMode().then(logResponse);
  } else if (id === "leave") {
    log(`About to execute [cnc.getChannel('1').cncServices.leaveBlockMode()]`);
    channel.cncServices.leaveBlockMode().then(logResponse);
  } else if (id === "send") {
    const val = $("#block").val();
    if (val) {
      log(
        `About to execute [cnc.getChannel('1').cncServices.setBlock(${JSON.stringify(
          val
        )})]`
      );
      channel.cncServices.setBlock(val).then(logResponse);
    }
  }
});

$(".jog").on("click", async function () {
  const id = $(this).attr("id");

  if (id === "start") {
    await conn.sendJogKeysToCnc("start", true);
    await new Promise((resolve) => setTimeout(() => resolve()), 1000);
    await conn.sendJogKeysToCnc("start", false);
  }

  if (id === "stop") {
    await conn.sendJogKeysToCnc("stop", true);
    await new Promise((resolve) => setTimeout(() => resolve()), 1000);
    await conn.sendJogKeysToCnc("stop", false);
  }

  if (id === "reset") {
    await conn.sendJogKeysToCnc("reset", true);
    await new Promise((resolve) => setTimeout(() => resolve()), 1000);
    await conn.sendJogKeysToCnc("reset", false);
  }
});

let listenerInstatiated = false;
let listener;
$("#chn1TableReactive button.start").on("click", function () {
  $("#chn1TableReactive button.start").prop("disabled", true);
  $("#chn1TableReactive button.stop").prop("disabled", false);
  if (!listenerInstatiated) {
    listener = cnc
      .getChannel(1)
      .activeToolProperties.createListener(["t", "d"]);
    log("Creating dynamic listener for activeTools props t & d");
    listenerInstatiated = true;
  }

  listener.start(function (activeTool, props) {
    log(`Updating ${props.join(",")} on active tool`);

    props.map((prop) => {
      const value = activeTool[prop];
      const $container = $("#chn1TableReactive .container-" + prop);
      $container.html(value).css("background-color", "pink");
      setTimeout(() => $container.css("background-color", ""), 3000);
    });
  });
});

$("#chn1TableReactive button.stop").on("click", function () {
  $("#chn1TableReactive button.start").prop("disabled", false);
  $("#chn1TableReactive button.stop").prop("disabled", true);
  listener.stop();
});

// Handle pannels toggle links
$(".toggler").on("click", function (ev) {
  ev.preventDefault();
  $("#" + $(this).data("target")).slideToggle();
});

log("App Initialization");

function log(message) {
  $("pre")[0].prepend(new Date().toLocaleTimeString() + " · " + message + "\n");
}

function updateBlockModeButtons(mode) {
  if (mode === "BLK_SELEC") {
    $("#enter").attr("disabled", "disabled");
    $("#leave").attr("disabled", null);
    $("#send").attr("disabled", null);
  } else {
    $("#enter").attr("disabled", null);
    $("#leave").attr("disabled", "disabled");
    $("#send").attr("disabled", "disabled");
  }
}

function logResponse(r) {
  log(`Executed [${r.name}] with response (${JSON.stringify(r.returnCode)})`);
}

//Function that fills currentAxes array and generate table with the axes passed by argument.
const updateAxes = (axes) => {
  $tbody.html("");
  currentAxes.length = 0;

  axes.map((ax) => {
    //Empty position.
    if (ax.logicId == 0) {
      return;
    }

    currentAxes.push(ax.logicId);

    $tbody.append(
      $(
        `<tr id="ax${ax.logicId}">` +
          `<td class="logicId">${ax.logicId}</td>` +
          `<td class="name">${ax.name}</td>` +
          `<td class="atippos">${ax.atippos.external.toFixed(3)}</td>` +
          `<td class="home">${ax.pos.external.toFixed(3)}</td>` +
          `<td class="parked">${ax.parked}</td>` +
          "</tr>"
      )
    );
  });
};

//Function that fills currentSpindles array and generate table with the spindles passed by argument.
const updateSpindles = (spindles) => {
  $tSpindlesbody.html("");
  currentSpindles.length = 0;

  spindles.map((sp) => {
    //Empty position.
    if (sp.logicId == 0) {
      return;
    }

    currentSpindles.push(sp.logicId);

    $tSpindlesbody.append(
      $(
        `<tr id="sp${sp.logicId}">` +
          `<td class="logicId">${sp.logicId}</td>` +
          `<td class="name">${sp.name}</td>` +
          `<td class="pos">${sp.pos.external.toFixed(3)}</td>` +
          `<td class="sreal">${sp.sreal.external.toFixed(3)}</td>` +
          `<td class="parked">${sp.parked}</td>` +
          "</tr>"
      )
    );
  });
};

//Function that subscribes cnc to axes and spindles updated event. Cnc object is immutable, so
//we must take in account not listening the same event more than once.
const subscribeToAxesSpindlesUpdates = () => {
  if (!axesUpdatesSubscribed) {
    //Channel's axes array can be changed, so we have to update axis table.
    cnc.getChannel(1).on("axesUpdated", updateAxes);
  }

  if (!spindlesUpdatesSubscribed) {
    //Channel's axes array can be changed, so we have to update axis table.
    cnc.getChannel(1).on("spindlesUpdated", updateSpindles);
  }
};

function cleanString(input) {
  var output = "";
  for (var i = 0; i < input.length; i++) {
    if (input.charCodeAt(i) <= 127) {
      output += input.charAt(i);
    }
  }
  return output;
}
function decodeFromBase64(input) {
  input = input.replace(/\s/g, "");
  const padding = "=".repeat((4 - (input.length % 4)) % 4);
  const base64 = (input + padding).replace(/-/g, "+").replace(/_/g, "/");
  return base64;
}
function imgError(image) {
  console.log("error");
  image.onerror = "";
  image.src = "/images/noimage.gif";
  return true;
}
const b64toBlob = async (base64, type = "application/octet-stream") => {
  const res = await fetch(`${base64}`);
  return res.blob();
};

const connectGraphics = async () => {
  $("#graphics").keydown(async function (event) {
    switch (event.key) {
      case "+":
      case "-":
        cnc.graphics.zoomCamera(
          "lfc_poc",
          "lfc_poc_camera1",
          event.keyCode
        );
        break;
      case "ArrowLeft":
      case "ArrowRight":
      case "ArrowUp":
      case "ArrowDown": {
        event.preventDefault();
        event.stopPropagation();
        let rotateCode = 0;
        switch (event.key) {
          case "ArrowLeft":
            rotateCode = 4;
            break;
          case "ArrowRight":
            rotateCode = 6;
            break;
          case "ArrowUp":
            rotateCode = 8;
            break;
          case "ArrowDown":
            rotateCode = 2;
            break;
        }
        cnc.graphics.rotateCamera("lfc_poc", "lfc_poc_camera1", rotateCode);
      }
    }
  });
  const resp = await cnc.graphics.addClient("lfc_poc");
  if (resp.returnCode === 0 || resp.returnCode === -2) {
    cnc.graphics.setViewConfiguration("lfc_poc", "", {
      enhancePartEdges: true,
      viewAxes: true,
      viewPart: true,
      viewPath: true,
      viewToolColors: true,
    });
    cnc.graphics.setContextDimensions("lfc_poc", 860, 430, 3);
    cnc.graphics.setBackgroundColor("lfc_poc", { r: 0, g: 0, b: 0 });
    cnc.graphics.alignCamera("lfc_poc", "");
    let nErrors = 0, nLoads = 0;
    const resp = await cnc.graphics.connect("lfc_poc", {
      uuid: "u1",
      streamType: "sse",
      binaryType: "string",
      imageQuality: 100,
      imageType: 2,
      callbackFrame: async (frame) => {
        const $graphicsContainer = $("#graphics-container");
        $graphicsContainer.attr("src", frame.data);
        $graphicsContainer.attr("width", frame.width);
        $graphicsContainer.attr("height", frame.height);
        $graphicsContainer.attr("onerror", () => {
          console.log("error", nErrors);
          //nErrors++;
          //console.log(frame.data);
        });
        $graphicsContainer.attr("onload", () => {
          console.log("onload", nLoads);
          //nLoads++;
          //console.log(frame.data);
        });
      },
      callbackError: () => {},
    });
  }
};

//Function that subscribes cnc to axes and spindles updated event. Cnc object is immutable, so
//we must take in account not listening the same event more than once.
const unsubscribeAxesSpindlesUpdates = () => {
  if (axesUpdatesSubscribed) {
    //Channel's axes array can be changed, so we have to update axis table.
    cnc.getChannel(1).removeAllListeners();
  }

  if (!spindlesUpdatesSubscribed) {
    //Channel's axes array can be changed, so we have to update axis table.
    cnc.getChannel(1).removeAllListeners();
  }
};

//Client does not know the cnc state when connection is stablished, in order to
//get the real state and mode of cnc kernel we make a CNC_STATUS request to retriev this data.
const requestCncStatusAndPaintIt = () => {
  if (cnc.modelCreated) {
    cnc
      .getChannel(1)
      .cncServices.request("CNC_STATUS")
      .then((req) => {
        if (req.returnCode === 0) {
          $(".mode").html(req.data.mode);
          updateBlockModeButtons(req.data.mode);
          $(".state").html(req.data.state);
        }
      });
  }
};
